
def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%d.g%s" % (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%d.g%s" % (pieces["distance"],
                                          pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_branch(pieces):
    """TAG[[.dev0]+DISTANCE.gHEX[.dirty]] .

    The ".dev0" means not master branch. Note that .dev0 sorts backwards
    (a feature branch will appear "older" than the master branch).

    Exceptions:
    1: no tags. 0[.dev0]+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            if pieces["branch"] != "master":
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "%d.g%s" % (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0"
        if pieces["branch"] != "master":
            rendered += ".dev0"
        rendered += "+untagged.%d.g%s" % (pieces["distance"],
                                          pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def pep440_split_post(ver):
    """Split pep440 version string at the post-release segment.

    Returns the release segments before the post-release and the
    post-release version number (or -1 if no post-release segment is present).
    """
    vc = str.split(ver, ".post")
    return vc[0], int(vc[1] or 0) if len(vc) == 2 else None


def render_pep440_pre(pieces):
    """TAG[.postN.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post0.devDISTANCE
    """
    if pieces["closest-tag"]:
        if pieces["distance"]:
            # update the post release segment
            tag_version, post_version = pep440_split_post(pieces["closest-tag"])
            rendered = tag_version
            if post_version is not None:
                rendered += ".post%d.dev%d" % (post_version+1, pieces["distance"])
            else:
                rendered += ".post0.dev%d" % (pieces["distance"])
        else:
            # no commits, use the tag as the version
            rendered = pieces["closest-tag"]
    else:
        # exception #1
        rendered = "0.post0.dev%d" % pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%s" % pieces["short"]
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
    return rendered


def render_pep440_post_branch(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX[.dirty]] .

    The ".dev0" means not master branch.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]+gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["branch"] != "master":
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%s" % pieces["short"]
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["branch"] != "master":
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {"version": "unknown",
                "full-revisionid": pieces.get("long"),
                "dirty": None,
                "error": pieces["error"],
                "date": None}

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-branch":
        rendered = render_pep440_branch(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-post-branch":
        rendered = render_pep440_post_branch(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%s'" % style)

    return {"version": rendered, "full-revisionid": pieces["long"],
            "dirty": pieces["dirty"], "error": None,
            "date": pieces.get("date")}




import sys # --STRIP DURING BUILD
def run_command(): pass # --STRIP DURING BUILD

def do_vcs_install(manifest_in, versionfile_source, ipy):
    """Git-specific installation logic for Versioneer.

    For Git, this means creating/changing .gitattributes to mark _version.py
    for export-subst keyword substitution.
    """
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    files = [manifest_in, versionfile_source]
    if ipy:
        files.append(ipy)
    try:
        my_path = __file__
        if my_path.endswith(".pyc") or my_path.endswith(".pyo"):
            my_path = os.path.splitext(my_path)[0] + ".py"
        versioneer_file = os.path.relpath(my_path)
    except NameError:
        versioneer_file = "versioneer.py"
    files.append(versioneer_file)
    present = False
    try:
        with open(".gitattributes", "r") as fobj:
            for line in fobj:
                if line.strip().startswith(versionfile_source):
                    if "export-subst" in line.strip().split()[1:]:
                        present = True
                        break
    except OSError:
        pass
    if not present:
        with open(".gitattributes", "a+") as fobj:
            fobj.write(f"{versionfile_source} export-subst\n")
        files.append(".gitattributes")
    run_command(GITS, ["add", "--"] + files)





import os, tempfile, unittest
from versioneer import versions_from_file

class Parser(unittest.TestCase):
    def test_lf(self):
        with tempfile.TemporaryDirectory() as root:
            fn = os.path.join(root, "_version.py")
            with open(fn, "wb") as f:
                f.write(b"version_json = '''\n{}\n'''  # END VERSION_JSON\n")
            data = versions_from_file(fn)
            self.assertEqual(data, {})

    def test_crlf(self):
        with tempfile.TemporaryDirectory() as root:
            fn = os.path.join(root, "_version.py")
            with open(fn, "wb") as f:
                f.write(b"version_json = '''\r\n{}\r\n'''  # END VERSION_JSON\r\n")
            data = versions_from_file(fn)
            self.assertEqual(data, {})




import os, tempfile
from setuptools import setup
from distutils.command.build_scripts import build_scripts
import versioneer
commands = versioneer.get_cmdclass().copy()

class my_build_scripts(build_scripts):
    def run(self):
        versions = versioneer.get_versions()
        tempdir = tempfile.mkdtemp()
        generated = os.path.join(tempdir, "rundemo")
        with open(generated, "wb") as f:
            for line in open("src/rundemo-template", "rb"):
                if line.strip().decode("ascii") == "versions = None":
                    f.write(('versions = %r\n' % (versions,)).encode("ascii"))
                else:
                    f.write(line)
        self.scripts = [generated]
        rc = build_scripts.run(self)
        os.unlink(generated)
        os.rmdir(tempdir)
        return rc
commands["build_scripts"] = my_build_scripts

setup(name="demo",
      version=versioneer.get_version(),
      description="Demo",
      url="url",
      author="author",
      author_email="email",
      zip_safe=True,
      scripts=["src/dummy"], # this will be replaced by my_build_scripts
      # without py_modules= or packages=, distutils thinks this module is not
      # "pure", and will put a platform indicator in the .whl name even
      # though we call bdist_wheel with --universal.
      py_modules=["fake"],
      cmdclass=commands,
      )



#!/usr/bin/env python

import os, sys, base64

VERSIONEER_b64 = """
@VERSIONEER-INSTALLER@
"""
newver = "@VERSIONEER-VERSION@"


def main():
    if len(sys.argv) < 2:
        print("Usage: versioneer install")
        sys.exit(1)

    command = sys.argv[1]
    if command in ("version", "--version"):
        print("versioneer (installer) %s" % newver)
        sys.exit(0)

    if command != "install":
        print("Usage: versioneer install")
        sys.exit(1)

    v = base64.b64decode(VERSIONEER_b64.encode('ASCII'))
    if os.path.exists("versioneer.py"):
        for line in open("versioneer.py").readlines()[:5]:
            if line.startswith("# Version: "):
                oldver = line[len("# Version: "):].strip()
                if oldver != newver:
                    print("replacing old versioneer.py (%s)" % oldver)
                break
        else:
            print("overwriting existing versioneer.py")
    with open("versioneer.py", "wb") as f:
        f.write(v)
    print("versioneer.py (%s) installed into local tree" % newver)
    print("Now running 'versioneer.py setup' to install the generated files..")
    if " " in sys.executable:
        import subprocess  # Do not import unless absolutely necessary to avoid compatibility issues
        subprocess.call([sys.executable, "versioneer.py", "setup"])
    else:
        os.execl(sys.executable, sys.executable, "versioneer.py", "setup")

if __name__ == '__main__':
    main()



import sys  # --STRIP DURING BUILD
import re  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
def register_vcs_handler(*args):  # --STRIP DURING BUILD
    def nil(f):  # --STRIP DURING BUILD
        return f  # --STRIP DURING BUILD
    return nil  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
def run_command(): pass  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
  # --STRIP DURING BUILD
class NotThisMethod(Exception):  # --STRIP DURING BUILD
    pass  # --STRIP DURING BUILD
@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, runner=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    GITS = ["git"]
    TAG_PREFIX_REGEX = "*"
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
        TAG_PREFIX_REGEX = r"\*"

    _, rc = runner(GITS, ["rev-parse", "--git-dir"], cwd=root,
                   hide_stderr=True)
    if rc != 0:
        if verbose:
            print("Directory %s not under git control" % root)
        raise NotThisMethod("'git rev-parse --git-dir' returned error")

    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out, rc = runner(GITS, ["describe", "--tags", "--dirty",
                                     "--always", "--long",
                                     "--match",
                                     "%s%s" % (tag_prefix, TAG_PREFIX_REGEX)],
                              cwd=root)
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out, rc = runner(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    branch_name, rc = runner(GITS, ["rev-parse", "--abbrev-ref", "HEAD"],
                             cwd=root)
    # --abbrev-ref was added in git-1.6.3
    if rc != 0 or branch_name is None:
        raise NotThisMethod("'git rev-parse --abbrev-ref' returned error")
    branch_name = branch_name.strip()

    if branch_name == "HEAD":
        # If we aren't exactly on a branch, pick a branch which represents
        # the current commit. If all else fails, we are on a branchless
        # commit.
        branches, rc = runner(GITS, ["branch", "--contains"], cwd=root)
        # --contains was added in git-1.5.4
        if rc != 0 or branches is None:
            raise NotThisMethod("'git branch --contains' returned error")
        branches = branches.split("\n")

        # Remove the first line if we're running detached
        if "(" in branches[0]:
            branches.pop(0)

        # Strip off the leading "* " from the list of branches.
        branches = [branch[2:] for branch in branches]
        if "master" in branches:
            branch_name = "master"
        elif not branches:
            branch_name = None
        else:
            # Pick the first branch that is returned. Good or bad.
            branch_name = branches[0]

    pieces["branch"] = branch_name

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[:git_describe.rindex("-dirty")]

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r'^(.+)-(\d+)-g([0-9a-f]+)$', git_describe)
        if not mo:
            # unparsable. Maybe git-describe is misbehaving?
            pieces["error"] = ("unable to parse git-describe output: '%s'"
                               % describe_out)
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%s' doesn't start with prefix '%s'"
                print(fmt % (full_tag, tag_prefix))
            pieces["error"] = ("tag '%s' doesn't start with prefix '%s'"
                               % (full_tag, tag_prefix))
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix):]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out, rc = runner(GITS, ["rev-list", "HEAD", "--count"], cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    # commit date: see ISO-8601 comment in git_versions_from_keywords()
    date = runner(GITS, ["show", "-s", "--format=%ci", "HEAD"], cwd=root)[0].strip()
    # Use only the last line.  Previous lines may contain GPG signature
    # information.
    date = date.splitlines()[-1]
    pieces["date"] = date.strip().replace(" ", "T", 1).replace(" ", "", 1)

    return pieces




import unittest
import os, tempfile
import configparser

from versioneer import get_config_from_root

base = """
[versioneer]
VCS = git
style = pep440
versionfile_source = petmail/_version.py
versionfile_build = petmail/_version.py
tag_prefix = v
parentdir_prefix = petmail-
"""

class Parser(unittest.TestCase):
    def parse(self, contents):
        with tempfile.TemporaryDirectory() as root:
            with open(os.path.join(root, "setup.cfg"), "w") as f:
                f.write(contents)
            return get_config_from_root(root)

    def test_base(self):
        cfg = self.parse(base)
        self.assertEqual(cfg.VCS, "git")
        self.assertEqual(cfg.style, "pep440")
        self.assertEqual(cfg.versionfile_source, "petmail/_version.py")
        self.assertEqual(cfg.versionfile_build, "petmail/_version.py")
        self.assertEqual(cfg.tag_prefix, "v")
        self.assertEqual(cfg.parentdir_prefix, "petmail-")
        self.assertEqual(cfg.verbose, None)

    def test_empty(self):
        self.assertRaises(configparser.NoSectionError,
                          self.parse, "")

    def test_mostly_empty(self):
        self.assertRaises(configparser.NoOptionError,
                          self.parse, "[versioneer]\n")

    def test_minimal(self):
        cfg = self.parse("[versioneer]\nvcs = git\n")
        self.assertEqual(cfg.VCS, "git")
        self.assertEqual(cfg.style, "")
        self.assertEqual(cfg.versionfile_source, None)
        self.assertEqual(cfg.versionfile_build, None)
        self.assertEqual(cfg.tag_prefix, None)
        self.assertEqual(cfg.parentdir_prefix, None)
        self.assertEqual(cfg.verbose, None)

    def test_empty_tag_prefixes(self):
        # all three of these should give an empty tag_prefix:
        #  tag_prefix =
        #  tag_prefix = ''
        #  tag_prefix = ""
        cfg = self.parse("[versioneer]\nVCS=git\ntag_prefix=")
        self.assertEqual(cfg.tag_prefix, "")
        cfg = self.parse("[versioneer]\nVCS=git\ntag_prefix=''")
        self.assertEqual(cfg.tag_prefix, "")
        cfg = self.parse("[versioneer]\nVCS=git\ntag_prefix=\"\"")
        self.assertEqual(cfg.tag_prefix, "")



import os, sys # --STRIP DURING BUILD
def get_root(): pass # --STRIP DURING BUILD
def get_config_from_root(): pass # --STRIP DURING BUILD
def versions_from_file(): pass # --STRIP DURING BUILD
def versions_from_parentdir(): pass # --STRIP DURING BUILD
def render(): pass # --STRIP DURING BUILD
HANDLERS = {} # --STRIP DURING BUILD
class NotThisMethod(Exception): pass  # --STRIP DURING BUILD

class VersioneerBadRootError(Exception):
    """The project root directory is unknown or missing key files."""


def get_versions(verbose=False):
    """Get the project version from whatever source is available.

    Returns dict with two keys: 'version' and 'full'.
    """
    if "versioneer" in sys.modules:
        # see the discussion in cmdclass.py:get_cmdclass()
        del sys.modules["versioneer"]

    root = get_root()
    cfg = get_config_from_root(root)

    assert cfg.VCS is not None, "please set [versioneer]VCS= in setup.cfg"
    handlers = HANDLERS.get(cfg.VCS)
    assert handlers, "unrecognized VCS '%s'" % cfg.VCS
    verbose = verbose or cfg.verbose
    assert cfg.versionfile_source is not None, \
        "please set versioneer.versionfile_source"
    assert cfg.tag_prefix is not None, "please set versioneer.tag_prefix"

    versionfile_abs = os.path.join(root, cfg.versionfile_source)

    # extract version from first of: _version.py, VCS command (e.g. 'git
    # describe'), parentdir. This is meant to work for developers using a
    # source checkout, for users of a tarball created by 'setup.py sdist',
    # and for users of a tarball/zipball created by 'git archive' or github's
    # download-from-tag feature or the equivalent in other VCSes.

    get_keywords_f = handlers.get("get_keywords")
    from_keywords_f = handlers.get("keywords")
    if get_keywords_f and from_keywords_f:
        try:
            keywords = get_keywords_f(versionfile_abs)
            ver = from_keywords_f(keywords, cfg.tag_prefix, verbose)
            if verbose:
                print("got version from expanded keyword %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        ver = versions_from_file(versionfile_abs)
        if verbose:
            print("got version from file %s %s" % (versionfile_abs, ver))
        return ver
    except NotThisMethod:
        pass

    from_vcs_f = handlers.get("pieces_from_vcs")
    if from_vcs_f:
        try:
            pieces = from_vcs_f(cfg.tag_prefix, root, verbose)
            ver = render(pieces, cfg.style)
            if verbose:
                print("got version from VCS %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        if cfg.parentdir_prefix:
            ver = versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
            if verbose:
                print("got version from parentdir %s" % ver)
            return ver
    except NotThisMethod:
        pass

    if verbose:
        print("unable to compute version")

    return {"version": "0+unknown", "full-revisionid": None,
            "dirty": None, "error": "unable to compute version",
            "date": None}


def get_version():
    """Get the short version string for this project."""
    return get_versions()["version"]



import os, sys, shutil, unittest, tempfile, tarfile, virtualenv, warnings
from wheel.bdist_wheel import get_abi_tag, get_platform
from packaging.tags import interpreter_name, interpreter_version

sys.path.insert(0, "src")
from from_file import versions_from_file
import common


pyver_major = "py%d" % sys.version_info[0]
pyver = "py%d.%d" % sys.version_info[:2]


# For binary wheels with native code
impl, impl_ver = interpreter_name(), interpreter_version()
abi = get_abi_tag()
try:
    plat = get_platform(None)
except TypeError:  # wheel < 0.34
    plat = get_platform()


class _Invocations(common.Common):
    def setUp(self):
        if False:
            # when debugging, put the generated files in a predictable place
            self.testdir = os.path.abspath("t")
            if os.path.exists(self.testdir):
                return
            os.mkdir(self.testdir)
        else:
            self.testdir = tempfile.mkdtemp()
        os.mkdir(self.subpath("cache"))
        os.mkdir(self.subpath("cache", "distutils"))
        os.mkdir(self.subpath("cache", "setuptools"))
        self.gitdir = None
        self.projdir = None

    def make_venv(self, mode):
        if not os.path.exists(self.subpath("venvs")):
            os.mkdir(self.subpath("venvs"))
        venv_dir = self.subpath("venvs/%s" % mode)
        # python3 on OS-X uses a funky two-part executable and an environment
        # variable to communicate between them. If this variable is still set
        # by the time a virtualenv's 'pip' or 'python' is run, and if that
        # command spawns another sys.executable underneath it, that second
        # child may use the wrong python, and can install things into the
        # real system library instead of the virtualenv. Invoking
        # virtualenv.create_environment() clears this as a side-effect, but
        # to make things safe I'll just clear this now. See
        # https://github.com/pypa/virtualenv/issues/322 and
        # https://bugs.python.org/issue22490 for some hints. I tried
        # switching to 'venv' on py3, but only py3.4 includes pip, and even
        # then it's an ancient version.
        os.environ.pop("__PYVENV_LAUNCHER__", None)
        virtualenv.logger = virtualenv.Logger([]) # hush
        # virtualenv causes DeprecationWarning/ResourceWarning on py3
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            virtualenv.create_environment(venv_dir)
            self.run_in_venv(venv_dir, venv_dir,
                             'pip', 'install', '-U',
                             'pip', 'wheel', 'packaging')
        return venv_dir

    def run_in_venv(self, venv, workdir, command, *args):
        bins = {"python": os.path.join(venv, "bin", "python"),
                "pip": os.path.join(venv, "bin", "pip"),
                "rundemo": os.path.join(venv, "bin", "rundemo"),
                }
        if command == "pip":
            args = ["--isolated", "--no-cache-dir"] + list(args)
        return self.command(bins[command], *args, workdir=workdir)

    def check_in_venv(self, venv):
        out = self.run_in_venv(venv, venv, "rundemo")
        v = dict(line.split(":", 1) for line in out.splitlines())
        self.assertEqual(v["version"], "2.0")
        return v

    def check_in_venv_withlib(self, venv):
        v = self.check_in_venv(venv)
        self.assertEqual(v["demolib"], "1.0")

    # "demolib" has a version of 1.0 and is built with distutils
    # "demoapp2-distutils" is v2.0, uses distutils, and has no deps
    # "demoapp2-setuptools" is v2.0, uses setuptools, and depends on demolib

    # repos and unpacked git-archive tarballs come in two flavors: normal (in
    # which the setup.py/setup.cfg/versioneer.py files live in the root of
    # the source tree), and "subproject" (where they live in a subdirectory).
    # sdists are always "normal" (although they might have come from either
    # normal or subproject -style source trees), and wheels/eggs don't have
    # these files at all.

    # TODO: git-archive subproject-flavor

    def make_demolib_sdist(self):
        # create an sdist of demolib-1.0 . for the *lib*, we only use the
        # tarball, never the repo.
        demolib_sdist = self.subpath("cache", "demolib-1.0.tar")
        if os.path.exists(demolib_sdist):
            return demolib_sdist
        libdir = self.subpath("build-demolib")
        shutil.copytree("test/demolib", libdir)
        shutil.copy("versioneer.py", libdir)
        self.git("init", workdir=libdir)
        self.python("versioneer.py", "setup", workdir=libdir)
        self.git("add", "--all", workdir=libdir)
        self.git("commit", "-m", "comment", workdir=libdir)
        self.git("tag", "demolib-1.0", workdir=libdir)
        self.python("setup.py", "sdist", "--format=tar", workdir=libdir)
        created = os.path.join(libdir, "dist", "demolib-1.0.tar")
        self.assertTrue(os.path.exists(created))
        shutil.copyfile(created, demolib_sdist)
        return demolib_sdist

    def make_linkdir(self):
        # create/populate a fake pypi directory for use with --find-links
        linkdir = self.subpath("linkdir")
        if os.path.exists(linkdir):
            return linkdir
        os.mkdir(linkdir)
        demolib_sdist = self.make_demolib_sdist()
        shutil.copy(demolib_sdist, linkdir)
        return linkdir

    def make_empty_indexdir(self):
        indexdir = self.subpath("indexdir")
        if os.path.exists(indexdir):
            return indexdir
        os.mkdir(indexdir)
        return indexdir

    def make_distutils_repo(self):
        # create a clean repo of demoapp2-distutils at 2.0
        repodir = self.subpath("demoapp2-distutils-repo")
        if os.path.exists(repodir):
            shutil.rmtree(repodir)
        shutil.copytree("test/demoapp2-distutils", repodir)
        shutil.copy("versioneer.py", repodir)
        self.git("init", workdir=repodir)
        self.python("versioneer.py", "setup", workdir=repodir)
        self.git("add", "--all", workdir=repodir)
        self.git("commit", "-m", "comment", workdir=repodir)
        self.git("tag", "demoapp2-2.0", workdir=repodir)
        return repodir

    def make_distutils_repo_subproject(self):
        # create a clean repo of demoapp2-distutils at 2.0
        repodir = self.subpath("demoapp2-distutils-repo-subproject")
        if os.path.exists(repodir):
            shutil.rmtree(repodir)
        shutil.copytree("test/demoapp2-distutils-subproject", repodir)
        projectdir = os.path.join(repodir, "subproject")
        shutil.copy("versioneer.py", projectdir)
        self.git("init", workdir=repodir)
        self.python("versioneer.py", "setup", workdir=projectdir)
        self.git("add", "--all", workdir=repodir)
        self.git("commit", "-m", "comment", workdir=repodir)
        self.git("tag", "demoapp2-2.0", workdir=repodir)
        return projectdir

    def make_distutils_wheel_with_pip(self):
        # create an wheel of demoapp2-distutils at 2.0
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        demoapp2_distutils_wheel = self.subpath("cache", "distutils", wheelname)
        if os.path.exists(demoapp2_distutils_wheel):
            return demoapp2_distutils_wheel
        repodir = self.make_distutils_repo()
        venv = self.make_venv("make-distutils-wheel-with-pip")
        self.run_in_venv(venv, repodir,
                         "pip", "wheel", "--wheel-dir", "wheelhouse",
                         "--no-index",# "--find-links", linkdir,
                         ".")
        created = os.path.join(repodir, "wheelhouse", wheelname)
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_distutils_wheel)
        return demoapp2_distutils_wheel

    def make_distutils_sdist(self):
        # create an sdist tarball of demoapp2-distutils at 2.0
        demoapp2_distutils_sdist = self.subpath("cache", "distutils",
                                                "demoapp2-2.0.tar")
        if os.path.exists(demoapp2_distutils_sdist):
            return demoapp2_distutils_sdist
        repodir = self.make_distutils_repo()
        self.python("setup.py", "sdist", "--format=tar", workdir=repodir)
        created = os.path.join(repodir, "dist", "demoapp2-2.0.tar")
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_distutils_sdist)
        return demoapp2_distutils_sdist

    def make_distutils_sdist_subproject(self):
        demoapp2_distutils_sdist = self.subpath("cache", "distutils",
                                                "demoapp2-subproject-2.0.tar")
        if os.path.exists(demoapp2_distutils_sdist):
            return demoapp2_distutils_sdist
        projectdir = self.make_distutils_repo_subproject()
        self.python("setup.py", "sdist", "--format=tar", workdir=projectdir)
        created = os.path.join(projectdir, "dist", "demoapp2-2.0.tar")
        # if that gets the version wrong, it will make the wrong tarball, and
        # this check will fail
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_distutils_sdist)
        return demoapp2_distutils_sdist

    def make_distutils_unpacked(self):
        sdist = self.make_distutils_sdist()
        unpack_into = self.subpath("demoapp2-distutils-unpacked")
        if os.path.exists(unpack_into):
            shutil.rmtree(unpack_into)
        os.mkdir(unpack_into)
        with tarfile.TarFile(sdist) as t:
            t.extractall(path=unpack_into)
        unpacked = os.path.join(unpack_into, "demoapp2-2.0")
        self.assertTrue(os.path.exists(unpacked))
        return unpacked

    def make_distutils_subproject_unpacked(self):
        sdist = self.make_distutils_sdist_subproject()
        unpack_into = self.subpath("demoapp2-distutils-unpacked-subproject")
        if os.path.exists(unpack_into):
            shutil.rmtree(unpack_into)
        os.mkdir(unpack_into)
        with tarfile.TarFile(sdist) as t:
            t.extractall(path=unpack_into)
        unpacked = os.path.join(unpack_into, "demoapp2-2.0")
        self.assertTrue(os.path.exists(unpacked))
        return unpacked

    def make_setuptools_repo(self):
        # create a clean repo of demoapp2-setuptools at 2.0
        repodir = self.subpath("demoapp2-setuptools-repo")
        if os.path.exists(repodir):
            shutil.rmtree(repodir)
        shutil.copytree("test/demoapp2-setuptools", repodir)
        shutil.copy("versioneer.py", repodir)
        self.git("init", workdir=repodir)
        self.python("versioneer.py", "setup", workdir=repodir)
        self.git("add", "--all", workdir=repodir)
        self.git("commit", "-m", "comment", workdir=repodir)
        self.git("tag", "demoapp2-2.0", workdir=repodir)
        return repodir

    def make_setuptools_extension_unpacked(self):
        sdist = self.make_setuptools_extension_sdist()
        unpack_into = self.subpath("demoappext-setuptools-unpacked")
        if os.path.exists(unpack_into):
            shutil.rmtree(unpack_into)
        os.mkdir(unpack_into)
        with tarfile.TarFile(sdist) as t:
            t.extractall(path=unpack_into)
        unpacked = os.path.join(unpack_into, "demoappext-2.0")
        self.assertTrue(os.path.exists(unpacked))
        return unpacked

    def make_setuptools_extension_sdist(self):
        # create an sdist tarball of demoappext-setuptools at 2.0
        demoappext_setuptools_sdist = self.subpath("cache", "setuptools",
                                                   "demoappext-2.0.tar")
        if os.path.exists(demoappext_setuptools_sdist):
            return demoappext_setuptools_sdist
        repodir = self.make_setuptools_extension_repo()
        self.python("setup.py", "sdist", "--format=tar", workdir=repodir)
        created = os.path.join(repodir, "dist", "demoappext-2.0.tar")
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoappext_setuptools_sdist)
        return demoappext_setuptools_sdist

    def make_setuptools_extension_repo(self):
        # create a clean repo of demoappext-setuptools at 2.0
        repodir = self.subpath("demoappext-setuptools-repo")
        if os.path.exists(repodir):
            shutil.rmtree(repodir)
        # import ipdb; ipdb.set_trace()
        shutil.copytree("test/demoappext-setuptools", repodir)
        shutil.copy("versioneer.py", repodir)
        self.git("init", workdir=repodir)
        self.python("versioneer.py", "setup", workdir=repodir)
        self.git("add", "--all", workdir=repodir)
        self.git("commit", "-m", "comment", workdir=repodir)
        self.git("tag", "demoappext-2.0", workdir=repodir)
        return repodir

    def make_setuptools_repo_subproject(self):
        # create a clean repo of demoapp2-setuptools at 2.0
        repodir = self.subpath("demoapp2-setuptools-repo-subproject")
        if os.path.exists(repodir):
            shutil.rmtree(repodir)
        shutil.copytree("test/demoapp2-setuptools-subproject", repodir)
        projectdir = os.path.join(repodir, "subproject")
        shutil.copy("versioneer.py", projectdir)
        self.git("init", workdir=repodir)
        self.python("versioneer.py", "setup", workdir=projectdir)
        self.git("add", "--all", workdir=repodir)
        self.git("commit", "-m", "comment", workdir=repodir)
        self.git("tag", "demoapp2-2.0", workdir=repodir)
        return projectdir

    def make_setuptools_sdist(self):
        # create an sdist tarball of demoapp2-setuptools at 2.0
        demoapp2_setuptools_sdist = self.subpath("cache", "setuptools",
                                                 "demoapp2-2.0.tar")
        if os.path.exists(demoapp2_setuptools_sdist):
            return demoapp2_setuptools_sdist
        repodir = self.make_setuptools_repo()
        self.python("setup.py", "sdist", "--format=tar", workdir=repodir)
        created = os.path.join(repodir, "dist", "demoapp2-2.0.tar")
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_setuptools_sdist)
        return demoapp2_setuptools_sdist

    def make_setuptools_sdist_subproject(self):
        demoapp2_setuptools_sdist = self.subpath("cache", "setuptools",
                                                 "demoapp2-subproject-2.0.tar")
        if os.path.exists(demoapp2_setuptools_sdist):
            return demoapp2_setuptools_sdist
        projectdir = self.make_setuptools_repo_subproject()
        self.python("setup.py", "sdist", "--format=tar", workdir=projectdir)
        created = os.path.join(projectdir, "dist", "demoapp2-2.0.tar")
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_setuptools_sdist)
        return demoapp2_setuptools_sdist

    def make_setuptools_unpacked(self):
        sdist = self.make_setuptools_sdist()
        unpack_into = self.subpath("demoapp2-setuptools-unpacked")
        if os.path.exists(unpack_into):
            shutil.rmtree(unpack_into)
        os.mkdir(unpack_into)
        with tarfile.TarFile(sdist) as t:
            t.extractall(path=unpack_into)
        unpacked = os.path.join(unpack_into, "demoapp2-2.0")
        self.assertTrue(os.path.exists(unpacked))
        return unpacked

    def make_setuptools_subproject_unpacked(self):
        sdist = self.make_setuptools_sdist_subproject()
        unpack_into = self.subpath("demoapp2-setuptools-unpacked-subproject")
        if os.path.exists(unpack_into):
            shutil.rmtree(unpack_into)
        os.mkdir(unpack_into)
        with tarfile.TarFile(sdist) as t:
            t.extractall(path=unpack_into)
        unpacked = os.path.join(unpack_into, "demoapp2-2.0")
        self.assertTrue(os.path.exists(unpacked))
        return unpacked

    def make_setuptools_egg(self):
        # create an egg of demoapp2-setuptools at 2.0
        demoapp2_setuptools_egg = self.subpath("cache", "setuptools",
                                               "demoapp2-2.0-%s.egg" % pyver)
        if os.path.exists(demoapp2_setuptools_egg):
            return demoapp2_setuptools_egg
        repodir = self.make_setuptools_repo()
        self.python("setup.py", "bdist_egg", workdir=repodir)
        created = os.path.join(repodir, "dist", "demoapp2-2.0-%s.egg" % pyver)
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_setuptools_egg)
        return demoapp2_setuptools_egg

    def make_setuptools_wheel_with_setup_py(self):
        # create an wheel of demoapp2-setuptools at 2.0
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        demoapp2_setuptools_wheel = self.subpath("cache", "setuptools",
                                                 wheelname)
        if os.path.exists(demoapp2_setuptools_wheel):
            # there are two ways to make this .whl, and we need to exercise
            # both, so don't actually cache the results
            os.unlink(demoapp2_setuptools_wheel)
        repodir = self.make_setuptools_repo()
        self.python("setup.py", "bdist_wheel", workdir=repodir)
        created = os.path.join(repodir, "dist", wheelname)
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_setuptools_wheel)
        return demoapp2_setuptools_wheel

    def make_setuptools_wheel_with_pip(self):
        # create an wheel of demoapp2-setuptools at 2.0
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        demoapp2_setuptools_wheel = self.subpath("cache", "setuptools",
                                                 wheelname)
        if os.path.exists(demoapp2_setuptools_wheel):
            # there are two ways to make this .whl, and we need to exercise
            # both, so don't actually cache the results
            os.unlink(demoapp2_setuptools_wheel)
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_repo()
        venv = self.make_venv("make-setuptools-wheel-with-pip")
        self.run_in_venv(venv, repodir,
                         "pip", "wheel", "--wheel-dir", "wheelhouse",
                         "--no-index", "--find-links", linkdir,
                         ".")
        created = os.path.join(repodir, "wheelhouse", wheelname)
        self.assertTrue(os.path.exists(created), created)
        shutil.copyfile(created, demoapp2_setuptools_wheel)
        return demoapp2_setuptools_wheel

    def make_binary_wheelname(self, app):
        return "%s-2.0-%s-%s-%s.whl" % (app,
            "".join([impl, impl_ver]), abi, plat.replace("-", "_"))


class DistutilsRepo(_Invocations, unittest.TestCase):
    def test_build(self):
        repodir = self.make_distutils_repo()
        self.python("setup.py", "build", workdir=repodir)
        # test that the built _version.py is correct. Ideally we'd actually
        # run PYTHONPATH=.../build/lib build/scripts-PYVER/rundemo and check
        # the output, but that's more fragile than I want to deal with today
        fn = os.path.join(repodir, "build", "lib", "demo", "_version.py")
        data = versions_from_file(fn)
        self.assertEqual(data["version"], "2.0")

    def test_install(self):
        repodir = self.make_distutils_repo()
        venv = self.make_venv("distutils-repo-install")
        self.run_in_venv(venv, repodir, "python", "setup.py", "install")
        self.check_in_venv(venv)

    def test_install_subproject(self):
        projectdir = self.make_distutils_repo_subproject()
        venv = self.make_venv("distutils-repo-install-subproject")
        self.run_in_venv(venv, projectdir, "python", "setup.py", "install")
        self.check_in_venv(venv)

    def test_pip_wheel(self):
        self.make_distutils_wheel_with_pip()
        # asserts version as a side-effect

    def test_sdist(self):
        sdist = self.make_distutils_sdist() # asserts version as a side-effect
        # make sure we used distutils/sdist, not setuptools/sdist
        with tarfile.TarFile(sdist) as t:
            self.assertFalse("demoapp2-2.0/src/demoapp2.egg-info/PKG-INFO" in
                             t.getnames())

    def test_sdist_subproject(self):
        sdist = self.make_distutils_sdist_subproject()
        # make sure we used distutils/sdist, not setuptools/sdist
        with tarfile.TarFile(sdist) as t:
            self.assertFalse("demoapp2-2.0/src/demoapp2.egg-info/PKG-INFO" in
                             t.getnames())

    def test_pip_install(self):
        repodir = self.make_distutils_repo()
        venv = self.make_venv("distutils-repo-pip-install")
        self.run_in_venv(venv, repodir, "pip", "install", ".")
        self.check_in_venv(venv)

    def test_pip_install_subproject(self):
        projectdir = self.make_distutils_repo_subproject()
        venv = self.make_venv("distutils-repo-pip-install-subproject")
        self.run_in_venv(venv, projectdir, "pip", "install", ".")
        self.check_in_venv(venv)

    def test_pip_install_from_afar(self):
        repodir = self.make_distutils_repo()
        venv = self.make_venv("distutils-repo-pip-install-from-afar")
        self.run_in_venv(venv, venv, "pip", "install", repodir)
        self.check_in_venv(venv)

    def test_pip_install_from_afar_subproject(self):
        projectdir = self.make_distutils_repo_subproject()
        venv = self.make_venv("distutils-repo-pip-install-from-afar-subproject")
        self.run_in_venv(venv, venv, "pip", "install", projectdir)
        self.check_in_venv(venv)

    def test_pip_install_editable(self):
        repodir = self.make_distutils_repo()
        venv = self.make_venv("distutils-repo-pip-install-editable")
        self.run_in_venv(venv, repodir, "pip", "install", "--editable", ".")
        self.check_in_venv(venv)

    def test_pip_install_editable_subproject(self):
        projectdir = self.make_distutils_repo_subproject()
        venv = self.make_venv("distutils-repo-pip-install-editable-subproject")
        self.run_in_venv(venv, projectdir, "pip", "install", "--editable", ".")
        self.check_in_venv(venv)

class SetuptoolsRepo(_Invocations, unittest.TestCase):
    def test_install(self):
        repodir = self.make_setuptools_repo()
        demolib = self.make_demolib_sdist()
        venv = self.make_venv("setuptools-repo-install")
        # "setup.py install" doesn't take --no-index or --find-links, so we
        # pre-install the dependency
        self.run_in_venv(venv, venv, "pip", "install", demolib)
        self.run_in_venv(venv, repodir, "python", "setup.py", "install")
        self.check_in_venv_withlib(venv)

    def test_install_subproject(self):
        projectdir = self.make_setuptools_repo_subproject()
        demolib = self.make_demolib_sdist()
        venv = self.make_venv("setuptools-repo-install-subproject")
        # "setup.py install" doesn't take --no-index or --find-links, so we
        # pre-install the dependency
        self.run_in_venv(venv, venv, "pip", "install", demolib)
        self.run_in_venv(venv, projectdir, "python", "setup.py", "install")
        self.check_in_venv_withlib(venv)

    def test_develop(self):
        linkdir = self.make_linkdir()
        indexdir = self.make_empty_indexdir()
        repodir = self.make_setuptools_repo()
        venv = self.make_venv("setuptools-repo-develop")
        # "setup.py develop" takes --find-links and --index-url but not
        # --no-index
        self.run_in_venv(venv, repodir,
                         "python", "setup.py", "develop",
                         "--index-url", indexdir, "--find-links", linkdir,
                         )
        self.check_in_venv_withlib(venv)

    def test_develop_subproject(self):
        linkdir = self.make_linkdir()
        indexdir = self.make_empty_indexdir()
        projectdir = self.make_setuptools_repo_subproject()
        venv = self.make_venv("setuptools-repo-develop-subproject")
        # "setup.py develop" takes --find-links and --index-url but not
        # --no-index
        self.run_in_venv(venv, projectdir,
                         "python", "setup.py", "develop",
                         "--index-url", indexdir, "--find-links", linkdir,
                         )
        self.check_in_venv_withlib(venv)

    def test_egg(self):
        self.make_setuptools_egg() # asserts version as a side-effect

    def test_pip_wheel(self):
        self.make_setuptools_wheel_with_pip()
        # asserts version as a side-effect

    def test_bdist_wheel(self):
        self.make_setuptools_wheel_with_setup_py()
        # asserts version as a side-effect

    def test_sdist(self):
        sdist = self.make_setuptools_sdist() # asserts version as a side-effect
        # make sure we used setuptools/sdist, not distutils/sdist
        with tarfile.TarFile(sdist) as t:
            self.assertIn("demoapp2-2.0/src/demoapp2.egg-info/PKG-INFO", t.getnames())

    def test_sdist_subproject(self):
        sdist = self.make_setuptools_sdist_subproject()
        # make sure we used setuptools/sdist, not distutils/sdist
        with tarfile.TarFile(sdist) as t:
            self.assertIn("demoapp2-2.0/src/demoapp2.egg-info/PKG-INFO", t.getnames())

    def test_pip_install(self):
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_repo()
        venv = self.make_venv("setuptools-repo-pip-install")
        self.run_in_venv(venv, repodir, "pip", "install", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_subproject(self):
        linkdir = self.make_linkdir()
        projectdir = self.make_setuptools_repo_subproject()
        venv = self.make_venv("setuptools-repo-pip-install-subproject")
        self.run_in_venv(venv, projectdir, "pip", "install", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_from_afar(self):
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_repo()
        venv = self.make_venv("setuptools-repo-pip-install-from-afar")
        self.run_in_venv(venv, venv, "pip", "install", repodir,
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_from_afar_subproject(self):
        linkdir = self.make_linkdir()
        projectdir = self.make_setuptools_repo_subproject()
        venv = self.make_venv("setuptools-repo-pip-install-from-afar-subproject")
        self.run_in_venv(venv, venv, "pip", "install", projectdir,
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_editable(self):
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_repo()
        venv = self.make_venv("setuptools-repo-pip-install-editable")
        self.run_in_venv(venv, repodir, "pip", "install", "--editable", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_editable_subproject(self):
        linkdir = self.make_linkdir()
        projectdir = self.make_setuptools_repo_subproject()
        venv = self.make_venv("setuptools-repo-pip-install-editable-subproject")
        self.run_in_venv(venv, projectdir, "pip", "install", "--editable", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

class DistutilsSdist(_Invocations, unittest.TestCase):
    def test_pip_install(self):
        sdist = self.make_distutils_sdist()
        venv = self.make_venv("distutils-sdist-pip-install")
        self.run_in_venv(venv, venv,
                         "pip", "install", sdist)
        self.check_in_venv(venv)

    def test_pip_install_subproject(self):
        sdist = self.make_distutils_sdist_subproject()
        venv = self.make_venv("distutils-sdist-pip-install-subproject")
        self.run_in_venv(venv, venv,
                         "pip", "install", sdist)
        self.check_in_venv(venv)

class SetuptoolsSdist(_Invocations, unittest.TestCase):
    def test_pip_install(self):
        linkdir = self.make_linkdir()
        sdist = self.make_setuptools_sdist()
        venv = self.make_venv("setuptools-sdist-pip-install")
        self.run_in_venv(venv, venv,
                         "pip", "install",
                         "--no-index", "--find-links", linkdir,
                         sdist)
        self.check_in_venv_withlib(venv)

    def test_pip_install_subproject(self):
        linkdir = self.make_linkdir()
        sdist = self.make_setuptools_sdist_subproject()
        venv = self.make_venv("setuptools-sdist-pip-install-subproject")
        self.run_in_venv(venv, venv,
                         "pip", "install",
                         "--no-index", "--find-links", linkdir,
                         sdist)
        self.check_in_venv_withlib(venv)

class SetuptoolsWheel(_Invocations, unittest.TestCase):
    def test_pip_install(self):
        linkdir = self.make_linkdir()
        wheel = self.make_setuptools_wheel_with_setup_py()
        venv = self.make_venv("setuptools-wheel-pip-install")
        self.run_in_venv(venv, venv,
                         "pip", "install",
                         "--no-index", "--find-links", linkdir,
                         wheel)
        self.check_in_venv_withlib(venv)

class DistutilsUnpacked(_Invocations, unittest.TestCase):
    def test_build(self):
        unpacked = self.make_distutils_unpacked()
        self.python("setup.py", "build", workdir=unpacked)
        # test that the built _version.py is correct. Ideally we'd actually
        # run PYTHONPATH=.../build/lib build/scripts-PYVER/rundemo and check
        # the output, but that's more fragile than I want to deal with today
        fn = os.path.join(unpacked, "build", "lib", "demo", "_version.py")
        data = versions_from_file(fn)
        self.assertEqual(data["version"], "2.0")

    def test_install(self):
        unpacked = self.make_distutils_unpacked()
        venv = self.make_venv("distutils-unpacked-install")
        self.run_in_venv(venv, unpacked, "python", "setup.py", "install")
        self.check_in_venv(venv)

    def test_install_subproject(self):
        unpacked = self.make_distutils_subproject_unpacked()
        venv = self.make_venv("distutils-subproject-unpacked-install")
        self.run_in_venv(venv, unpacked, "python", "setup.py", "install")
        self.check_in_venv(venv)

    def test_pip_wheel(self):
        unpacked = self.make_distutils_unpacked()
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        venv = self.make_venv("distutils-unpacked-pip-wheel")
        self.run_in_venv(venv, unpacked,
                         "pip", "wheel", "--wheel-dir", "wheelhouse",
                         "--no-index",# "--find-links", linkdir,
                         ".")
        created = os.path.join(unpacked, "wheelhouse", wheelname)
        self.assertTrue(os.path.exists(created), created)

    def test_pip_install(self):
        repodir = self.make_distutils_unpacked()
        venv = self.make_venv("distutils-unpacked-pip-install")
        self.run_in_venv(venv, repodir, "pip", "install", ".")
        self.check_in_venv(venv)

    def test_pip_install_subproject(self):
        unpacked = self.make_distutils_subproject_unpacked()
        venv = self.make_venv("distutils-subproject-unpacked-pip-install")
        self.run_in_venv(venv, unpacked, "pip", "install", ".")
        self.check_in_venv(venv)

    def test_pip_install_from_afar(self):
        repodir = self.make_distutils_unpacked()
        venv = self.make_venv("distutils-unpacked-pip-install-from-afar")
        self.run_in_venv(venv, venv, "pip", "install", repodir)
        self.check_in_venv(venv)

class SetuptoolsUnpacked(_Invocations, unittest.TestCase):
    def test_install(self):
        unpacked = self.make_setuptools_unpacked()
        demolib = self.make_demolib_sdist()
        venv = self.make_venv("setuptools-unpacked-install")
        # "setup.py install" doesn't take --no-index or --find-links, so we
        # pre-install the dependency
        self.run_in_venv(venv, venv, "pip", "install", demolib)
        self.run_in_venv(venv, unpacked,
                         "python", "setup.py", "install")
        self.check_in_venv_withlib(venv)

    def test_install_subproject(self):
        unpacked = self.make_setuptools_subproject_unpacked()
        demolib = self.make_demolib_sdist()
        venv = self.make_venv("setuptools-subproject-unpacked-install")
        # "setup.py install" doesn't take --no-index or --find-links, so we
        # pre-install the dependency
        self.run_in_venv(venv, venv, "pip", "install", demolib)
        self.run_in_venv(venv, unpacked,
                         "python", "setup.py", "install")
        self.check_in_venv_withlib(venv)

    def test_wheel(self):
        unpacked = self.make_setuptools_unpacked()
        self.python("setup.py", "bdist_wheel", workdir=unpacked)
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        wheel = os.path.join(unpacked, "dist", wheelname)
        self.assertTrue(os.path.exists(wheel))

    def test_pip_wheel(self):
        unpacked = self.make_setuptools_unpacked()
        linkdir = self.make_linkdir()
        wheelname = "demoapp2-2.0-%s-none-any.whl" % pyver_major
        venv = self.make_venv("setuptools-unpacked-pip-wheel")
        self.run_in_venv(venv, unpacked,
                         "pip", "wheel", "--wheel-dir", "wheelhouse",
                         "--no-index", "--find-links", linkdir,
                         ".")
        created = os.path.join(unpacked, "wheelhouse", wheelname)
        self.assertTrue(os.path.exists(created), created)

    def test_pip_install(self):
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_unpacked()
        venv = self.make_venv("setuptools-unpacked-pip-install")
        self.run_in_venv(venv, repodir, "pip", "install", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_subproject(self):
        linkdir = self.make_linkdir()
        unpacked = self.make_setuptools_subproject_unpacked()
        venv = self.make_venv("setuptools-subproject-unpacked-pip-install")
        self.run_in_venv(venv, unpacked, "pip", "install", ".",
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_pip_install_from_afar(self):
        linkdir = self.make_linkdir()
        repodir = self.make_setuptools_unpacked()
        venv = self.make_venv("setuptools-unpacked-pip-install-from-afar")
        self.run_in_venv(venv, venv, "pip", "install", repodir,
                         "--no-index", "--find-links", linkdir)
        self.check_in_venv_withlib(venv)

    def test_extension_wheel_setuptools(self):
        # create an wheel of demoappext-setuptools at 2.0
        wheelname = self.make_binary_wheelname('demoappext')
        demoappext_setuptools_wheel = self.subpath("cache", "setuptools",
                                                   wheelname)
        if os.path.exists(demoappext_setuptools_wheel):
            # there are two ways to make this .whl, and we need to exercise
            # both, so don't actually cache the results
            os.unlink(demoappext_setuptools_wheel)
        repodir = self.make_setuptools_extension_repo()
        self.python("setup.py", "bdist_wheel", workdir=repodir)
        created = os.path.join(repodir, "dist", wheelname)
        self.assertTrue(os.path.exists(created), created)

    def test_extension_inplace(self):
        # build extensions in place. No wheel package
        unpacked = self.make_setuptools_extension_unpacked()
        venv = self.make_venv("setuptools-unpacked-pip-wheel-extension")
        self.run_in_venv(venv, unpacked,
                         "python", "setup.py", "build_ext", "-i")
        # No wheel package is created, _version.py should exist in
        # module dir only
        version_file = os.path.join(unpacked, "demo", "_version.py")
        self.assertTrue(os.path.exists(version_file))

    def test_extension_wheel_pip(self):
        # create an wheel of demoappext-setuptools at 2.0 with pip
        wheelname = self.make_binary_wheelname('demoappext')
        demoappext_setuptools_wheel = self.subpath("cache", "setuptools",
                                                   wheelname)
        if os.path.exists(demoappext_setuptools_wheel):
            # there are two ways to make this .whl, and we need to exercise
            # both, so don't actually cache the results
            os.unlink(demoappext_setuptools_wheel)
        unpacked = self.make_setuptools_extension_unpacked()
        linkdir = self.make_linkdir()
        venv = self.make_venv("setuptools-unpacked-pip-wheel-extension")
        self.run_in_venv(venv, unpacked,
                         "pip", "wheel", "--wheel-dir", "wheelhouse",
                         "--no-index", "--find-links", linkdir,
                         ".")
        created = os.path.join(unpacked, "wheelhouse", wheelname)
        self.assertTrue(os.path.exists(created), created)


if __name__ == '__main__':
    unittest.main()




from setuptools import setup
import versioneer
commands = versioneer.get_cmdclass().copy()

setup(name="demoapp2",
      version=versioneer.get_version(),
      description="Demo",
      url="url",
      author="author",
      author_email="email",
      zip_safe=True,
      packages=["demo"],
      package_dir={"": "src"},
      entry_points={
          'console_scripts': [ 'rundemo = demo.main:run' ],
          },
      install_requires=["demolib==1.0"],
      cmdclass=commands,
      )



import unittest

from versioneer import render


class Testing_renderer_case_mixin(object):
    """
    This is a mixin object which can be combined with a unittest.TestCase
    which defines a style and an expected dictionary. See Test_pep440 for
    and example.

    """
    def define_pieces(self, closest_tag, distance=0, dirty=False, branch=False):
        return {"error": '',
                "closest-tag": closest_tag,
                "distance": distance,
                "dirty": dirty,
                "short": "abc" if distance else '',
                "long": "abcdefg" if distance else '',
                "date": "2016-05-31T13:02:11+0200",
                "branch": "feature" if branch else "master"}

    def assert_rendered(self, pieces, test_case_name):
        version = render(pieces, self.style)['version']
        expected = self.expected[test_case_name]
        msg = ('Versions differ for {0} style with "{1}" case: expected {2}, '
               'got {3}'.format(self.style, test_case_name, expected, version))
        self.assertEqual(version, expected, msg)

    # Naming structure:
    # test_(un)tagged_<n>_commits_(clean|dirty)
    def test_tagged_0_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3'),
                             'tagged_0_commits_clean')

    def test_tagged_1_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3', distance=1),
                             'tagged_1_commits_clean')

    def test_tagged_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3', dirty=True),
                             'tagged_0_commits_dirty')

    def test_tagged_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3', distance=1,
                                                dirty=True),
                             'tagged_1_commits_dirty')

    def test_untagged_0_commits_clean(self):
        self.assert_rendered(self.define_pieces(None),
                             'untagged_0_commits_clean')

    def test_untagged_1_commits_clean(self):
        self.assert_rendered(self.define_pieces(None, distance=1),
                             'untagged_1_commits_clean')

    def test_untagged_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces(None, dirty=True),
                             'untagged_0_commits_dirty')

    def test_untagged_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces(None, distance=1,
                                                dirty=True),
                             'untagged_1_commits_dirty')

    def test_error_getting_parts(self):
        self.assert_rendered({'error': 'Not a git repo'},
                             'error_getting_parts')


class Testing_branch_renderer_case_mixin(Testing_renderer_case_mixin):
    """
    This is a mixin object which extends the base mixin and adds tests
    that also test on the value of the branch in the dictionary.

    """

    # Naming structure:
    # test_branch_(un)tagged_<n>_commits_(clean|dirty)
    def test_branch_tagged_0_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3', branch=True),
                             'branch_tagged_0_commits_clean')

    def test_branch_tagged_1_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3', branch=True,
                                                distance=1),
                             'branch_tagged_1_commits_clean')

    def test_branch_tagged_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3', branch=True,
                                                dirty=True),
                             'branch_tagged_0_commits_dirty')

    def test_branch_tagged_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3', branch=True,
                                                distance=1, dirty=True),
                             'branch_tagged_1_commits_dirty')

    def test_branch_untagged_0_commits_clean(self):
        self.assert_rendered(self.define_pieces(None, branch=True),
                             'branch_untagged_0_commits_clean')

    def test_branch_untagged_1_commits_clean(self):
        self.assert_rendered(self.define_pieces(None, branch=True, distance=1),
                             'branch_untagged_1_commits_clean')

    def test_branch_untagged_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces(None, branch=True, dirty=True),
                             'branch_untagged_0_commits_dirty')

    def test_branch_untagged_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces(None, branch=True, distance=1,
                                                dirty=True),
                             'branch_untagged_1_commits_dirty')


class Testing_post_renderer_case_mixin(Testing_renderer_case_mixin):
    """
    This is a mixin object which extends the base mixin and adds tests
    that also test version tags with a post-release segment.

    """

    # Naming structure:
    # test_(un)tagged_post<n>_<n>_commits_(clean|dirty)
    def test_tagged_post_0_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post'),
                             'tagged_post_0_commits_clean')
    
    def test_tagged_post1_0_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post1'),
                             'tagged_post1_0_commits_clean')
    
    def test_tagged_post_1_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post', distance=1),
                             'tagged_post_1_commits_clean')
    
    def test_tagged_post1_1_commits_clean(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post1', distance=1),
                             'tagged_post1_1_commits_clean')

    def test_tagged_post_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post', dirty=True),
                             'tagged_post_0_commits_dirty')
    
    def test_tagged_post1_0_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post1', dirty=True),
                             'tagged_post1_0_commits_dirty')
    
    def test_tagged_post_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post', distance=1,
                                                dirty=True),
                             'tagged_post_1_commits_dirty')
    
    def test_tagged_post1_1_commits_dirty(self):
        self.assert_rendered(self.define_pieces('v1.2.3.post1', distance=1,
                                                dirty=True),
                             'tagged_post1_1_commits_dirty')


class Test_pep440(unittest.TestCase, Testing_renderer_case_mixin):
    style = 'pep440'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3+0.g.dirty',
                'tagged_1_commits_clean': 'v1.2.3+1.gabc',
                'tagged_1_commits_dirty': 'v1.2.3+1.gabc.dirty',
                'untagged_0_commits_clean': '0+untagged.0.g',
                'untagged_0_commits_dirty': '0+untagged.0.g.dirty',
                'untagged_1_commits_clean': '0+untagged.1.gabc',
                'untagged_1_commits_dirty': '0+untagged.1.gabc.dirty',
                'error_getting_parts': 'unknown'
                }


class Test_pep440_branch(unittest.TestCase,
                         Testing_branch_renderer_case_mixin):
    style = 'pep440-branch'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3+0.g.dirty',
                'tagged_1_commits_clean': 'v1.2.3+1.gabc',
                'tagged_1_commits_dirty': 'v1.2.3+1.gabc.dirty',
                'untagged_0_commits_clean': '0+untagged.0.g',
                'untagged_0_commits_dirty': '0+untagged.0.g.dirty',
                'untagged_1_commits_clean': '0+untagged.1.gabc',
                'untagged_1_commits_dirty': '0+untagged.1.gabc.dirty',
                'branch_tagged_0_commits_clean': 'v1.2.3',
                'branch_tagged_0_commits_dirty': 'v1.2.3.dev0+0.g.dirty',
                'branch_tagged_1_commits_clean': 'v1.2.3.dev0+1.gabc',
                'branch_tagged_1_commits_dirty': 'v1.2.3.dev0+1.gabc.dirty',
                'branch_untagged_0_commits_clean': '0.dev0+untagged.0.g',
                'branch_untagged_0_commits_dirty': '0.dev0+untagged.0.g.dirty',
                'branch_untagged_1_commits_clean': '0.dev0+untagged.1.gabc',
                'branch_untagged_1_commits_dirty': '0.dev0+untagged.1.gabc.dirty',
                'error_getting_parts': 'unknown'
                }


class Test_pep440_old(unittest.TestCase, Testing_renderer_case_mixin):
    style = 'pep440-old'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3.post0.dev0',
                'tagged_1_commits_clean': 'v1.2.3.post1',
                'tagged_1_commits_dirty': 'v1.2.3.post1.dev0',
                'untagged_0_commits_clean': '0.post0',
                'untagged_0_commits_dirty': '0.post0.dev0',
                'untagged_1_commits_clean': '0.post1',
                'untagged_1_commits_dirty': '0.post1.dev0',
                'error_getting_parts': 'unknown'
                }


class Test_pep440_post(unittest.TestCase, Testing_renderer_case_mixin):
    style = 'pep440-post'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3.post0.dev0+g',
                'tagged_1_commits_clean': 'v1.2.3.post1+gabc',
                'tagged_1_commits_dirty': 'v1.2.3.post1.dev0+gabc',
                'untagged_0_commits_clean': '0.post0+g',
                'untagged_0_commits_dirty': '0.post0.dev0+g',
                'untagged_1_commits_clean': '0.post1+gabc',
                'untagged_1_commits_dirty': '0.post1.dev0+gabc',
                'error_getting_parts': 'unknown'
                }


class Test_pep440_post_branch(unittest.TestCase,
                              Testing_branch_renderer_case_mixin):
    style = 'pep440-post-branch'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3.post0+g.dirty',
                'tagged_1_commits_clean': 'v1.2.3.post1+gabc',
                'tagged_1_commits_dirty': 'v1.2.3.post1+gabc.dirty',
                'untagged_0_commits_clean': '0.post0+g',
                'untagged_0_commits_dirty': '0.post0+g.dirty',
                'untagged_1_commits_clean': '0.post1+gabc',
                'untagged_1_commits_dirty': '0.post1+gabc.dirty',
                'branch_tagged_0_commits_clean': 'v1.2.3',
                'branch_tagged_0_commits_dirty': 'v1.2.3.post0.dev0+g.dirty',
                'branch_tagged_1_commits_clean': 'v1.2.3.post1.dev0+gabc',
                'branch_tagged_1_commits_dirty': 'v1.2.3.post1.dev0+gabc.dirty',
                'branch_untagged_0_commits_clean': '0.post0.dev0+g',
                'branch_untagged_0_commits_dirty': '0.post0.dev0+g.dirty',
                'branch_untagged_1_commits_clean': '0.post1.dev0+gabc',
                'branch_untagged_1_commits_dirty': '0.post1.dev0+gabc.dirty',
                'error_getting_parts': 'unknown'
                }


class Test_pep440_pre(unittest.TestCase, Testing_post_renderer_case_mixin):
    style = 'pep440-pre'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3',
                'tagged_1_commits_clean': 'v1.2.3.post0.dev1',
                'tagged_1_commits_dirty': 'v1.2.3.post0.dev1',
                'untagged_0_commits_clean': '0.post0.dev0',
                'untagged_0_commits_dirty': '0.post0.dev0',
                'untagged_1_commits_clean': '0.post0.dev1',
                'untagged_1_commits_dirty': '0.post0.dev1',
                'tagged_post_0_commits_clean': 'v1.2.3.post',
                'tagged_post1_0_commits_clean': 'v1.2.3.post1',
                'tagged_post_1_commits_clean': 'v1.2.3.post1.dev1',
                'tagged_post1_1_commits_clean': 'v1.2.3.post2.dev1',
                'tagged_post_0_commits_dirty': 'v1.2.3.post',
                'tagged_post1_0_commits_dirty': 'v1.2.3.post1',
                'tagged_post_1_commits_dirty': 'v1.2.3.post1.dev1',
                'tagged_post1_1_commits_dirty': 'v1.2.3.post2.dev1',
                'error_getting_parts': 'unknown'
                }


class Test_git_describe(unittest.TestCase, Testing_renderer_case_mixin):
    style = 'git-describe'
    expected = {'tagged_0_commits_clean': 'v1.2.3',
                'tagged_0_commits_dirty': 'v1.2.3-dirty',
                'tagged_1_commits_clean': 'v1.2.3-1-gabc',
                'tagged_1_commits_dirty': 'v1.2.3-1-gabc-dirty',
                'untagged_0_commits_clean': '',
                'untagged_0_commits_dirty': '-dirty',
                'untagged_1_commits_clean': 'abc',
                'untagged_1_commits_dirty': 'abc-dirty',
                'error_getting_parts': 'unknown'
                }


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import os, base64, tempfile, io
from pathlib import Path
from setuptools import setup, Command
from setuptools.command.build_py import build_py
from setuptools.dist import Distribution as _Distribution

LONG = Path.read_text(Path(__file__).parent / "README.md")

# as nice as it'd be to versioneer ourselves, that sounds messy.
VERSION = "0.21"


def ver(s):
    return s.replace("@VERSIONEER-VERSION@", VERSION)

def get(fn, add_ver=False, unquote=False, do_strip=False, do_readme=False):
    with open(fn) as f:
        text = f.read()

    if add_ver:
        text = ver(text)
    if unquote:
        text = text.replace("%", "%%")
    if do_strip:
        text = "".join(line for line in text.splitlines(keepends=True)
                         if not line.endswith("# --STRIP DURING BUILD\n"))
    if do_readme:
        text = text.replace("@README@", get("README.md"))
    return text

def get_vcs_list():
    project_path = Path(__file__).absolute().parent / "src"
    return [filename
            for filename
            in os.listdir(str(project_path))
            if Path.is_dir(project_path / filename) and filename != "__pycache__"]

def generate_long_version_py(VCS):
    s = io.StringIO()
    s.write(get(f"src/{VCS}/long_header.py", add_ver=True, do_strip=True))
    for piece in ["src/subprocess_helper.py",
                  "src/from_parentdir.py",
                  f"src/{VCS}/from_keywords.py",
                  f"src/{VCS}/from_vcs.py",
                  "src/render.py",
                  f"src/{VCS}/long_get_versions.py"]:
        s.write(get(piece, unquote=True, do_strip=True))
    return s.getvalue()

def generate_versioneer_py():
    s = io.StringIO()
    s.write(get("src/header.py", add_ver=True, do_readme=True))
    s.write(get("src/subprocess_helper.py", do_strip=True))

    for VCS in get_vcs_list():
        s.write(f"LONG_VERSION_PY['{VCS}'] = r'''\n")
        s.write(generate_long_version_py(VCS))
        s.write("'''\n")
        s.write("\n\n")

        s.write(get(f"src/{VCS}/from_keywords.py", do_strip=True))
        s.write(get(f"src/{VCS}/from_vcs.py", do_strip=True))

        s.write(get(f"src/{VCS}/install.py", do_strip=True))

    s.write(get("src/from_parentdir.py", do_strip=True))
    s.write(get("src/from_file.py", add_ver=True, do_strip=True))
    s.write(get("src/render.py", do_strip=True))
    s.write(get("src/get_versions.py", do_strip=True))
    s.write(get("src/cmdclass.py", do_strip=True))
    s.write(get("src/setupfunc.py", do_strip=True))

    return s.getvalue().encode("utf-8")


class make_versioneer(Command):
    description = "create standalone versioneer.py"
    user_options = []
    boolean_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        with open("versioneer.py", "w") as f:
            f.write(generate_versioneer_py().decode("utf8"))
        return 0

class make_long_version_py_git(Command):
    description = "create standalone _version.py (for git)"
    user_options = []
    boolean_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        assert os.path.exists("versioneer.py")
        long_version = generate_long_version_py("git")
        with open("git_version.py", "w") as f:
            f.write(long_version %
                    {"DOLLAR": "$",
                     "STYLE": "pep440",
                     "TAG_PREFIX": "tag-",
                     "PARENTDIR_PREFIX": "parentdir_prefix",
                     "VERSIONFILE_SOURCE": "versionfile_source",
                     })
        return 0

class my_build_py(build_py):
    def run(self):
        v = generate_versioneer_py()
        v_b64 = base64.b64encode(v).decode("ascii")
        lines = [v_b64[i:i+60] for i in range(0, len(v_b64), 60)]
        v_b64 = "\n".join(lines)+"\n"

        with open("src/installer.py") as f:
            s = f.read()
        s = ver(s.replace("@VERSIONEER-INSTALLER@", v_b64))

        with tempfile.TemporaryDirectory() as tempdir:
            installer = os.path.join(tempdir, "versioneer.py")
            with open(installer, "w") as f:
                f.write(s)

            self.py_modules = [os.path.splitext(os.path.basename(installer))[0]]
            self.package_dir.update({'': os.path.dirname(installer)})
            rc = build_py.run(self)
        return rc


# python's distutils treats module-less packages as binary-specific (not
# "pure"), so "setup.py bdist_wheel" creates binary-specific wheels. Override
# this so we get cross-platform wheels instead. More info at:
# https://bitbucket.org/pypa/wheel/issue/116/packages-with-only-filesdata_files-get
class Distribution(_Distribution):
    def is_pure(self): return True

setup(
    name = "versioneer",
    license = "public domain",
    version = VERSION,
    description = "Easy VCS-based management of project version strings",
    author = "Brian Warner",
    author_email = "warner-versioneer@lothar.com",
    url = "https://github.com/python-versioneer/python-versioneer",
    # "fake" is replaced with versioneer-installer in build_scripts. We need
    # a non-empty list to provoke "setup.py build" into making scripts,
    # otherwise it skips that step.
    py_modules = ["fake"],
    entry_points={
        'console_scripts': [
            'versioneer = versioneer:main',
        ],
    },
    long_description=LONG,
    long_description_content_type="text/markdown",
    distclass=Distribution,
    cmdclass = { "build_py": my_build_py,
                 "make_versioneer": make_versioneer,
                 "make_long_version_py_git": make_long_version_py_git,
                 },
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        ],
    )




from setuptools import setup
import versioneer
commands = versioneer.get_cmdclass().copy()

setup(name="demo",
      version=versioneer.get_version(),
      description="Demo",
      url="url",
      author="author",
      author_email="email",
      zip_safe=True,
      packages=["demo"],
      package_dir={"": "src"},
      scripts=["bin/rundemo"],
      cmdclass=commands,
      )
